/*
    BSFilter.cpp  

    Implementation of the BSFilter effect (Bit-Shift Filter).

    ---------------------------------------------------------------------------
 
    Copyright (C) 2016, Cornfield Electronics, Inc.
 
    This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 
    Unported License.
 
    To view a copy of this license, visit
    http://creativecommons.org/licenses/by-sa/3.0/
 
    Created by Bill Alessi & Mitch Altman.
 
   ---------------------------------------------------------------------------
*/

#include "types.h"
#include "Audio.h"
#include "Console_.h"
#include "BSFilter.h"

BSFilter::BSFilter() 
{ 
   shortcut = 'b';
   reset();
}

void BSFilter::charEv( char code )
{
   char digit;

   switch ( code )
   {
      case 'c':                        // set # bits to clip 

         digit = console.getDigit( "clip", MaxClip );
         if ( digit )
            setClip( digit - '0' );
         break;

      case 's':                        // set # bits to shift 

         digit = console.getDigit( "shift", 7 );
         if ( digit )
         {
            if ( digit < '1' ) digit = '1';
            setShift( digit - '0' );
         }
         break;

      case '-':                        // complement the filter output

         complement = true;
         break;

      case '+':                        // normal filter output

         complement = false;
         break;

      default:

         Effect::charEv( code );
   }
}

void BSFilter::info()
{
   Effect::info();

   if ( complement )
      console.print( '-' );
   else
      console.print( '+' );

   console.print( " shift: " );
   console.print( (int )shift );

   console.print( " clip: " );
   console.print( (int )numClip );
}

char BSFilter::menu( key k )
{
   switch ( k.position() )
   {
      case  2: return 's';
      case  4: return 'c';
      case  5: return '+';
      case  7: return '-';
      default: return Effect::menu( k );
   }
}

void BSFilter::process( char *buf )
{
   Int r;                        // working register for performing shifts

   byte icnt = audio::bufSz;     // process this many ticks of input

   while ( icnt-- )
   {
      r.val    = last * mltShift;
      r._.msb &= clipMask;
      r.val   -= last;
      r.val   += *buf;
      r.val  >>= shift;
      last     = r.val;

      if ( complement )
         r.val = (int )(*buf - last);

      *buf++ = r.val;
   }
}

char *BSFilter::prompt()
{
   return "bsf";
}

void BSFilter::reset()
{
   Effect::reset();
   setShift(1);                        // 50% exponential average
   setClip(0);                         // clip no bits
   complement = false;                 // normal filter output
}

void BSFilter::setClip( byte n )
{
   Word m;                            // working register for clipMask
   numClip  = n;
   m.val    = (word )(1 << (8 - n));
   m.val   -= 1;
   clipMask = m._.lsb;
}

void BSFilter::setShift( byte n )
{
   shift = n;
   mltShift = 1;
   while ( n ) 
   {
      mltShift += mltShift;
      --n;
   }
}



