/*
    LFO.cpp  

    Implementation of the LFO class.

    ---------------------------------------------------------------------------
 
    Copyright (C) 2016, Cornfield Electronics, Inc.
 
    This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 
    Unported License.
 
    To view a copy of this license, visit
    http://creativecommons.org/licenses/by-sa/3.0/
 
    Created by Bill Alessi & Mitch Altman.
 
   ---------------------------------------------------------------------------
*/

#include "Console_.h"
#include "LFO.h"
                        /*--------------------
                                LFO
                         --------------------*/

#define MAX_OSC_DEPTH   1.0      // maximum depth of oscillation
#define MIN_OSC_DEPTH   0.0      // minimum depth of oscillation
#define DEF_OSC_DEPTH   0.25     // default depth of oscillation
#define MAX_OSC_FREQ    6.0      // maximum freq  of oscillation
#define MIN_OSC_FREQ    0.01     // minimum freq  of oscillation
#define DEF_OSC_FREQ    3.8      // default freq  of oscillation

void LFO::calcRange()
{
   top = depth;
   bot = -depth;
}

void LFO::calcStep()
{
   extern double dynaRate;
   step = 2.0 * freq * (top - bot) / dynaRate;
}

void LFO::dynamics()
{
   if ( amMute() )
      val = 1.0;
   else
   {
      pos += step;               // bump oscillator position

      // turn oscillator back if bounds exceeded

      if ( pos > top )           
      {
         pos  = 2*top - pos;
         step = -step;
      }
      else if ( pos < bot )
      {
         pos  = 2*bot - pos;
         step = -step;
      }

      evaluate();                // set val
   }
}

void LFO::evaluate()
{
   val = pos;
}

void LFO::info()
{
   Control::info();
   console.infoDouble( "freq", freq );
   console.infoDouble( "depth", depth );
}

void LFO::charEv( char key )
{
   switch ( key )
   {
      case 'd':                  // set depth of oscillation

         if ( console.getDouble( "depth" ) )
            setDepth( lastDouble );
         break;

      case 'f':                  // set frequency of oscillation

         if ( console.getDouble( "freq" ) )
            setFreq( lastDouble );
         break;

      default:

         Control::charEv( key );
   }
}

char LFO::menu( key k )
{
   switch ( k.position() )
   {
      case  2: return 'd';
      case  5: return 'f';
      default: return Control::menu( k );
   }
}

char* LFO::prompt()
{
   return "LFO";
}

void LFO::reset()
{
   Control::reset();
   val   = 1.0;
   pos   = 0.0;
   setOsc( DEF_OSC_DEPTH, DEF_OSC_FREQ );
}

void LFO::setDepth( double d )
{
   if ( d > MAX_OSC_DEPTH )
      d = MAX_OSC_DEPTH;
   else if ( d < MIN_OSC_DEPTH )
      d = MIN_OSC_DEPTH;
   depth = d;
   calcRange();
   calcStep();
}

void LFO::setFreq( double f )
{
   if ( f > MAX_OSC_FREQ )
      f = MAX_OSC_FREQ;
   else if ( f < MIN_OSC_FREQ )
      f = MIN_OSC_FREQ;
   freq = f;
   calcStep();
}

void LFO::setOsc( double d, double f )
{
   freq  = f;
   setDepth( d );
}
