/*
    StepSqnc.cpp  

    Implementation of the Step Sequencer.

    ---------------------------------------------------------------------------
 
    Copyright (C) 2016, Cornfield Electronics, Inc.
 
    This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 
    Unported License.
 
    To view a copy of this license, visit
    http://creativecommons.org/licenses/by-sa/3.0/
 
    Created by Bill Alessi & Mitch Altman.
 
   ---------------------------------------------------------------------------
*/

#include "Audio.h"
#include "ConKeyBrd.h"
#include "StepSqnc.h"

#define DefTempo     108.00      // default tempo
#define tokenEOS     255         // end-of-sequence token

class StepProgrammer : public ConKeyBrd
{
   private:

   byte duration;                // duration of current note

   public:

   byte   bufLen;                // length of sequence buffer
   byte  *sqnc;                  // buffer for compiled sequence records
   byte   idx;                   // current index into sqnc[] 

   void butEv( but b )
   {
      byte      butNum  = b.num();
      butEvType butType = b.type();

      switch ( butType )
      {
         case butTAP:            // bump keybrd up(but 1) - down(but 0) 1 octave 

            keybrd.bumpOctave( butNum );
            break;

         case butPRESS:          // bump duration on button 1 press

            if ( butNum == 1 )
            {
               ++duration;
               break;
            }

         default:                 

            ConKeyBrd::butEv( b );
      }
   }

   void charEv( char code )      // handle a character event
   {
      switch ( code )
      {
         case '.':               // bump duration
         case ' ':

            ++duration;
            break;

         default:                         
         
            ConKeyBrd::charEv( code );
      }
   }

   boolean keyEv( key k )        // handle a (piano) key event
   {
      
      ConKeyBrd::keyEv( k );     // echoes key to console

      if ( idx > 0 )             // compile duration for prior note   
      {
         sqnc[ idx++ ] = duration;  
      }

      if ( idx > bufLen - 3 )    // no space for another note + EOS
      {
         sqnc[ idx ] = tokenEOS;
         console.popMode();
         return true;
      }
      else
      {
         sqnc[ idx++ ] = k.val;  // compile new note
         duration      = 1;      // set its preliminary duration to 1
      }

      return false;              // allow parent mode to produce feedback
   }
   
   void onFocus( focus f )
   {
      if ( f == focusPOP )
      {
         if ( idx > 0 )          // compile duration for current note
             sqnc[ idx++ ] = duration; 
         sqnc[ idx ] = tokenEOS;
      }
   }

   char *prompt()
   {
      return CONSTR("sqnc");
   }

} stepProgrammer;

bool StepSqnc::_charEv( char code )
{
   bool handled = true;

   switch ( code )
   {
      case '[':                  // start sequencer playback

         start();
         break;

      case ']':                  // stop sequencer playback

         stop();
         break;

      case '|':                  // pause-resume sequencer playback

         if ( playback != pbOFF ) playback ^= pbPAUSED;
         break;

      case 'S':                  // step program via console keybrd

         stop();                 // insure sequencer is not playing

         stepProgrammer.bufLen = MaxLen;
         stepProgrammer.sqnc   = &this->sqnc[0];
         stepProgrammer.idx    = 0;
         console.pushMode( &stepProgrammer );

         break;

      case 't':                  // set sequencer tempo

         if ( console.getDouble( "tempo" ) )
            setTempo( lastDouble );
         break;

      case 'k':                  // swallow "keybrd" event if sequence is playing

         if ( playback == pbON  ) break;

      default:                   // event has not been handled

         handled = false;
   }

   return handled;
}

void StepSqnc::exeRec()
{
   byte token = sqnc[ idx++ ];
   if ( token == tokenEOS )      // if end-of-sequence
   {
      if ( idx > 1 )             //    if non-null sqnc, loop back to 1st rec
      {
         idx = 0;
         exeRec();
      }
      else                       //    else null sqnc, turn off playback
         stop();
   }
   else                          // else process non-EOS
   {
      subject->keyEv( (key )token );
      exeDC = sqnc[ idx++ ];
   }
}

void StepSqnc::info()
{
   console.rtab();
   console.print( playback & pbON ? '[' : ']' );
   console.infoDouble( "t", tempo );
}

void StepSqnc::cont()
{
   if ( playback == pbON )                      // continue sequencer playback
   {
      if ( beatDC <= audio::bufSz )             // process next beat
      {
         beatDC += ticksPerBeat - audio::bufSz; // reload beat downcounter
         if ( --exeDC == 0 )                    // if next record has arrived
            exeRec();                           // execute it
      }
      else                                      // next beat not yet arrived
         beatDC -= audio::bufSz;                // update beat downcounter
   }
}

void StepSqnc::reset()
{
   playback = pbOFF;
   sqnc[0] = tokenEOS;
   setTempo( DefTempo );
}

void StepSqnc::setTempo( double t )
{
   const double minTempo = (audioRate * 60.0) / (double )0xFFFE;  

   if ( t < minTempo )  // make sure ticksPerBeat fits in a word
      t = minTempo;

   tempo        = t;
   ticksPerBeat = ((double )ticksPerSec * 60.0) / t;
   
}

void StepSqnc::start()
{
   playback = pbON;              // set playback on
   idx = 0;                      // rewind to 1st record
   exeRec();                     // execute it
   beatDC = ticksPerBeat;        // set beatDC for next beat
}

void StepSqnc::stop()
{
   playback = pbOFF;
}

